#define ASCRIPT_DLL_MODULE
//-----------------------------------------------------------------------------
// AScript win32ole module
//-----------------------------------------------------------------------------
#include <windows.h>
#include "Module_win32ole.h"
#include "Expr.h"

AScript_BeginModule(win32ole)

OLECHAR *ToOleChar(const char *psz);
char *ToChar(const OLECHAR *bstr);

void ToVariant(VARIANT &var, const Value &value);
void ToValue(Environment &env, Value &value, const VARIANT &var);

// Class_Ole / Object_Ole
AScript_DeclareClass(Ole);

class Object_Ole : public Object {
public:
	class Callable : public ICallable {
	private:
		Object_Ole *_pObj;
		DISPID _dispid;
	public:
		inline Callable(Object_Ole *pObj) : _pObj(pObj) {}
		HRESULT Setup(const TCHAR *name);
		virtual Value Call(Environment &env, Signal sig, ContextExpr &contextExpr);
	};
private:
	IDispatch *_pDispatch;
	Callable _callable;
public:
	inline static Object_Ole *GetSelfObj(Context &context) {
		return dynamic_cast<Object_Ole *>(context.GetSelfObj());
	}
public:
	inline Object_Ole(Class *pClass) :
						Object(pClass), _pDispatch(NULL), _callable(this) {}
	inline Object_Ole(Class *pClass, IDispatch *pDispatch) :
						Object(pClass), _pDispatch(pDispatch), _callable(this) {}
	virtual ~Object_Ole();
	virtual Object *Clone() const;
	inline IDispatch *GetDispatch() { return _pDispatch; }
	bool Create(Signal sig, const TCHAR *progID);
	HRESULT GetDispID(const char *name, DISPID &dispid);
	virtual Value EvalGetter(Environment &env, Signal sig,
								const Symbol *pSymbol, bool &evaluatedFlag);
	virtual Value EvalSetter(Environment &env, Signal sig,
				const Symbol *pSymbol, const Value &value, bool &evaluatedFlag);
	virtual ICallable *GetCallable(Environment &env, Signal sig, const Symbol *pSymbol);
	virtual String ToString(Signal sig, bool exprFlag);
	static void SetError(Signal sig, HRESULT hr);
};

Object_Ole::~Object_Ole()
{
	_pDispatch->Release();
}

Object *Object_Ole::Clone() const
{
	return NULL;
}

bool Object_Ole::Create(Signal sig, const TCHAR *progID)
{
	CLSID clsid;
	OLECHAR *progID_W = ToOleChar(progID);
	::CLSIDFromProgID(progID_W, &clsid);
	HRESULT hr = ::CoCreateInstance(clsid, NULL,
					CLSCTX_INPROC_SERVER | CLSCTX_LOCAL_SERVER,
					IID_IDispatch, reinterpret_cast<LPVOID *>(&_pDispatch));
	::SysFreeString(progID_W);
	if (FAILED(hr)) {
		SetError(sig, hr);
		return false;
	}
	return true;
}

HRESULT Object_Ole::GetDispID(const char *name, DISPID &dispid)
{
	OLECHAR *wszName = ToOleChar(name);
	HRESULT hr = _pDispatch->GetIDsOfNames(IID_NULL,
							&wszName, 1, LOCALE_USER_DEFAULT, &dispid);
	::SysFreeString(wszName);
	return hr;
}

Value Object_Ole::EvalGetter(Environment &env, Signal sig,
								const Symbol *pSymbol, bool &evaluatedFlag)
{
	//::_tprintf(_T("EvalGetter(%s)\n"), pSymbol->GetName());
	DISPID dispid;
	do {
		HRESULT hr = GetDispID(pSymbol->GetName(), dispid);
		if (FAILED(hr)) {
			SetError(sig, hr);
			return Value::Null;
		}
	} while (0);
	evaluatedFlag = true;
	VARIANT var;
	do {
		DISPPARAMS dispParams = { NULL, NULL, 0, 0 };
		HRESULT hr = _pDispatch->Invoke(dispid, IID_NULL, LOCALE_USER_DEFAULT,
						DISPATCH_PROPERTYGET, &dispParams, &var, NULL, NULL);
		if (FAILED(hr)) {
			SetError(sig, hr);
			return Value::Null;
		}
	} while (0);
	Value result;
	ToValue(env, result, var);
	::VariantClear(&var);
	return result;
}

Value Object_Ole::EvalSetter(Environment &env, Signal sig,
				const Symbol *pSymbol, const Value &value, bool &evaluatedFlag)
{
	//::_tprintf(_T("EvalSetter(%s)\n"), pSymbol->GetName());
	evaluatedFlag = true;
	DISPID dispid;
	do {
		HRESULT hr = GetDispID(pSymbol->GetName(), dispid);
		if (FAILED(hr)) {
			SetError(sig, hr);
			return Value::Null;
		}
	} while (0);
	do {
		VARIANTARG varArgs[1];
		DISPID dispidNamedArgs[1] = { DISPID_PROPERTYPUT };
		ToVariant(varArgs[0], value);
		DISPPARAMS dispParams;
		dispParams.rgvarg = varArgs;
		dispParams.rgdispidNamedArgs = dispidNamedArgs;
		dispParams.cArgs = 1;
		dispParams.cNamedArgs = 1;
		HRESULT hr = _pDispatch->Invoke(dispid, IID_NULL, LOCALE_USER_DEFAULT,
							DISPATCH_PROPERTYPUT, &dispParams, NULL, NULL, NULL);
		::VariantClear(&varArgs[0]);
		if (FAILED(hr)) {
			SetError(sig, hr);
			return Value::Null;
		}
	} while (0);
	return value;
}

ICallable *Object_Ole::GetCallable(Environment &env, Signal sig, const Symbol *pSymbol)
{
	//::_tprintf(_T("GetCallable(%s)\n"), pSymbol->GetName());
	HRESULT hr = _callable.Setup(pSymbol->GetName());
	if (FAILED(hr)) {
		SetError(sig, hr);
		return NULL;
	}
	return &_callable;
}

String Object_Ole::ToString(Signal sig, bool exprFlag)
{
	String rtn;
	rtn += _T("<ole>");
	return rtn;
}

HRESULT Object_Ole::Callable::Setup(const TCHAR *name)
{
	return _pObj->GetDispID(name, _dispid);
}

Value Object_Ole::Callable::Call(Environment &env, Signal sig, ContextExpr &contextExpr)
{
	//IDispatch::GetTypeInfoCount
	//IDispatch::GetTypeInfo
	//ITypeInfo
	const ExprList &exprArgs = contextExpr.GetArgs();
	VARIANTARG *varArgs = NULL;
	size_t cArgs = exprArgs.size();
	DISPID *dispidNamedArgs = NULL;
	size_t cNamedArgs = 0;
	// http://winapi.freetechsecrets.com/ole/OLEIDispatchInvoke.htm
	if (cArgs > 0) {
		varArgs = new VARIANTARG[cArgs];
		foreach_const (ExprList, ppExpr, exprArgs) {
			const Expr *pExpr = *ppExpr;
			if (pExpr->IsDictAssign()) cNamedArgs++;
		}
		dispidNamedArgs = new DISPID[cNamedArgs];
		int i = 0;
		foreach_const (ExprList, ppExpr, exprArgs) {
			const Expr *pExpr = *ppExpr;
			if (!pExpr->IsDictAssign()) continue;
			const Expr_DictAssign *pExprDictAssign =
						dynamic_cast<const Expr_DictAssign *>(pExpr);
			Value valueKey = pExprDictAssign->GetKey(sig);
			if (sig.IsSignalled()) return Value::Null;
			Value value = pExprDictAssign->GetRight()->Exec(env, sig);
			if (sig.IsSignalled()) return Value::Null;
			dispidNamedArgs[i] = 0; // how to get dispid?
			ToVariant(varArgs[i], value);
			i++;
		}
		i = cArgs;
		foreach_const (ExprList, ppExpr, exprArgs) {
			const Expr *pExpr = *ppExpr;
			if (pExpr->IsDictAssign()) continue;
			i--;
			Value value = pExpr->Exec(env, sig);
			if (sig.IsSignalled()) return Value::Null;
			ToVariant(varArgs[i], value);
		}
	}
	DISPPARAMS dispParams;
	dispParams.rgvarg = varArgs;
	dispParams.cArgs = cArgs;
	dispParams.rgdispidNamedArgs = dispidNamedArgs;
	dispParams.cNamedArgs = cNamedArgs;
	VARIANT varResult;
	HRESULT hr = _pObj->GetDispatch()->Invoke(_dispid, IID_NULL, LOCALE_USER_DEFAULT,
			DISPATCH_METHOD | DISPATCH_PROPERTYGET, &dispParams, &varResult, NULL, NULL);
	for (size_t i = 0; i < cArgs; i++) {
		::VariantClear(&varArgs[i]);
	}
	delete[] varArgs;
	delete[] dispidNamedArgs;
	if (FAILED(hr)) {
		::VariantClear(&varResult);
		Object_Ole::SetError(sig, hr);
		return Value::Null;
	}
	Value result;
	ToValue(env, result, varResult);
	::VariantClear(&varResult);
	return result;
}

void Object_Ole::SetError(Signal sig, HRESULT hr)
{
	LPTSTR errMsg = NULL;
	::FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
				NULL, hr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
				(LPTSTR)&errMsg, 0, NULL);
	if (errMsg == NULL) {
		sig.SetError(ERR_SystemError, _T("unknown ComError"));
	} else {
		size_t len = ::_tcslen(errMsg);
		if (len > 0 && errMsg[len - 1] == _T('\n')) errMsg[len - 1] = _T('\0');
		sig.SetError(ERR_SystemError, _T("ComError %s"), errMsg);
		::LocalFree(errMsg);
	}
}

AScript_ImplementClass(Ole)
{
	//AScript_AssignMethod(Ole, match);
}

// obj = win32ole.new(progid:string):map
AScript_DeclareFunctionEx(new_, _T("new"))
{
	SetMode(RSLTMODE_Normal, MAP_On);
	DeclareArg(env, _T("progid"), VTYPE_String);
}

AScript_ImplementFunction(new_)
{
	Object_Ole *pObj = new Object_Ole(AScript_Class(Ole, env));
	if (!pObj->Create(sig, context.GetString(0))) {
		delete pObj;
		return Value::Null;
	}
	return Value(pObj, VTYPE_Object);
}

OLECHAR *ToOleChar(const char *psz)
{
	// cnt includes null-terminater
	int cnt = ::MultiByteToWideChar(CP_ACP, 0, psz, -1, NULL, 0);
	OLECHAR *bstr = ::SysAllocStringByteLen(NULL, cnt * sizeof(OLECHAR));
	::MultiByteToWideChar(CP_ACP, 0, psz, -1, bstr, cnt);
	return bstr;
}

char *ToChar(const OLECHAR *bstr)
{
	// cnt includes null-terminater
	int cnt = ::WideCharToMultiByte(CP_ACP, 0, bstr, -1, NULL, 0, NULL, NULL);
	char *psz = new char[cnt];
	::WideCharToMultiByte(CP_ACP, 0, bstr, -1, psz, cnt, NULL, NULL);
	return psz;
}

void ToVariant(VARIANT &var, const Value &value)
{
	::VariantInit(&var);
	if (value.IsNumber()) {
		Number num = value.GetNumber();
		if (static_cast<Number>(static_cast<long>(num)) == num) {
			var.vt = VT_I4;
			var.lVal = static_cast<long>(value.GetNumber());
		} else {
			var.vt = VT_R8;
			var.dblVal = value.GetNumber();
		}
	} else if (value.IsString()) {
		var.vt = VT_BSTR;
		var.bstrVal = ToOleChar(value.GetString());
	} else if (value.IsBoolean()) {
		var.vt = VT_BOOL;
		var.boolVal = value.GetBoolean()? VARIANT_TRUE : VARIANT_FALSE;
	} else {
	}
}

void ToValue(Environment &env, Value &value, const VARIANT &var)
{
	VARTYPE type = var.vt & VT_TYPEMASK;
	if (var.vt & VT_ARRAY) {
		var.byref;
	} else if (type == VT_I4) {
		value = Value(static_cast<Number>((var.vt & VT_BYREF)? *var.plVal : var.lVal));
	} else if (type == VT_UI1) {
		value = Value(static_cast<Number>((var.vt & VT_BYREF)? *var.pbVal : var.bVal));
	} else if (type == VT_I2) {
		value = Value(static_cast<Number>((var.vt & VT_BYREF)? *var.piVal : var.iVal));
	} else if (type == VT_R4) {
		value = Value(static_cast<Number>((var.vt & VT_BYREF)? *var.pfltVal : var.fltVal));
	} else if (type == VT_R8) {
		value = Value(static_cast<Number>((var.vt & VT_BYREF)? *var.pdblVal :var.dblVal));
	} else if (type == VT_BOOL) {
		value = Value(static_cast<Number>((var.vt & VT_BYREF)? *var.pboolVal : var.boolVal));
	} else if (type == VT_ERROR) {
		//value = Value(static_cast<Number>((var.vt & VT_BYREF)? *var.pscode : var.code));
	} else if (type == VT_CY) {
		//value = Value(static_cast<Number>((var.vt & VT_BYREF)? *var.pcyVal : var.cyVal));
	} else if (type == VT_DATE) {
		value = Value(static_cast<Number>((var.vt & VT_BYREF)? *var.pdate : var.date));
	} else if (type == VT_BSTR) {
		char *p = ToChar(var.bstrVal);
		value = Value(env, p);
		delete[] p;
	} else if (type == VT_UNKNOWN) {
		//value = Value(static_cast<Number>((var.vt & VT_BYREF)? *var.ppunkVal : var.punkVal));
	} else if (type == VT_DISPATCH) {
		IDispatch *pDispatch = (var.vt & VT_BYREF)? *var.ppdispVal : var.pdispVal;
		pDispatch->AddRef(); // prevent deletion by VariantClear()
		Object_Ole *pObj = new Object_Ole(AScript_Class(Ole, env), pDispatch);
		value = Value(pObj, VTYPE_Object);
	} else if (type == VT_VARIANT) {
		//value = Value(static_cast<Number>(*var.pvarVal));
	}
}

AScript_ModuleEntry()
{
	::CoInitialize(0);
	AScript_AssignFunctionEx(new_, _T("new"));
}

AScript_ModuleTerminate()
{
	::CoUninitialize();
}

AScript_EndModule(win32ole)

AScript_DLLModuleEntry(win32ole)
